""" CLI utilities for Publ """
# pylint:disable=too-many-arguments

import itertools
import logging
import os.path
import re
import time

import arrow
import click
import slugify
from flask.cli import AppGroup, with_appcontext
from pony import orm

from . import queries
from .config import config

LOGGER = logging.getLogger(__name__)

publ_cli = AppGroup('publ', short_help="Publ-specific commands")  # pylint:disable=invalid-name


@publ_cli.command('reindex', short_help="Reindex the content store")
@click.option('--quietly', '-q', 'quietly', is_flag=True, help="Quietly")
@click.option('--fresh', '-f', 'fresh', is_flag=True, help="Start with a fresh database")
@with_appcontext
def reindex_command(quietly, fresh):
    """ Forces a reindex of the content store.

    This is particularly useful to ensure that all content has been indexed
    before performing another action, such as sending out notifications.
    """
    from . import index, model

    if fresh:
        model.reset()

    spinner = itertools.cycle('|/-\\')

    index.scan_index(config.content_folder, False)
    while index.in_progress():
        if not quietly:
            qlen = index.queue_size() or ''
            print(f"\rIndexing... {next(spinner)} {qlen}        ", end='', flush=True)
        time.sleep(0.1)
    if not quietly:
        print("Done")


@publ_cli.command('token', short_help="Generate a bearer token")
@click.argument('identity')
@click.option('--scope', '-s', help="The token's permission scope")
@click.option('--lifetime', '-l', help="The token's lifetime (in seconds)", default=3600)
@with_appcontext
def token_command(identity, scope, lifetime):
    """ Generates a bearer token for use with external applications. """
    from . import tokens
    print(tokens.get_token(identity, int(lifetime), scope))


@publ_cli.command('normalize', short_help="Normalize entry filenames")
@click.argument('category', nargs=-1)
@click.option('--recurse', '-r', 'recurse', is_flag=True,
              help="Include subdirectories")
@click.option('--all', '-a', 'all_entries', is_flag=True,
              help="Apply to all entries, not just reachable ones")
@click.option('--dry-run', '-n', 'dry_run', is_flag=True,
              help="Show, but don't apply, changes")
@click.option('--format', '-f', 'format_str',
              help="Filename format to use",
              default="{date} {sid} {title}")
@click.option('--verbose', '-v', 'verbose', is_flag=True,
              help="Show detailed actions")
@with_appcontext
@orm.db_session
def normalize_command(category, recurse, dry_run, format_str, verbose, all_entries):
    """ Normalizes the filenames of content files based on a standardized format.

    This will only normalize entries which are already in the content index.

    If no categories are specified, it defaults to the root category. To include
    the root category in a list of other categories, use an empty string parameter,
    e.g.:

        flask publ normalize '' blog

    Available tokens for --format/-f:

        {date}    The entry's publish date, in YYYYMMDD format

        {time}    The entry's publish time, in HHMMSS format

        {id}      The entry's ID

        {status}  The entry's publish status

        {sid}     If the entry is reachable, the ID, otherwise the status

        {title}   The entry's title, normalized to filename-safe characters

        {slug}    The entry's slug text

        {type}    The entry's type
    """
    # pylint:disable=too-many-locals

    from .model import PublishStatus

    entries = queries.build_query({
        'category': category or '',
        'recurse': recurse,
        '_future': True,
        '_all': all_entries,
    })

    fname_slugify = slugify.UniqueSlugify(max_length=100, safe_chars='-.', separator=' ')

    for entry in entries:
        path = os.path.dirname(entry.file_path)
        basename, ext = os.path.splitext(os.path.basename(entry.file_path))

        status = PublishStatus(entry.status)

        eid = entry.id
        if status == PublishStatus.DRAFT:
            # Draft entries don't get a stable entry ID
            eid = status.name

        sid = entry.id if status in (PublishStatus.PUBLISHED,
                                     PublishStatus.HIDDEN,
                                     PublishStatus.SCHEDULED) else status.name

        date = arrow.get(entry.local_date)

        dest_basename = format_str.format(
            date=date.format('YYYYMMDD'),
            time=date.format('HHmmss'),
            id=eid,
            status=status.name,
            sid=sid,
            title=entry.title,
            slug=entry.slug_text,
            type=entry.entry_type).strip()
        dest_basename = re.sub(r' +', ' ', dest_basename)

        if dest_basename != basename:
            while True:
                # UniqueSlugify will bump the suffix until it doesn't collide
                dest_path = os.path.join(path, fname_slugify(dest_basename) + ext)
                if not os.path.exists(dest_path):
                    break

            if verbose:
                print(f'{entry.file_path} -> {dest_path}')

            if not os.path.isfile(entry.file_path):
                LOGGER.warning('File %s does not exist; is the index up-to-date?', entry.file_path)
            elif os.path.exists(dest_path):
                LOGGER.warning('File %s already exists', dest_path)
            elif not dry_run:
                try:
                    os.rename(entry.file_path, dest_path)
                except OSError:
                    LOGGER.exception('Error moving %s to %s', entry.file_path, dest_path)
                entry.file_path = dest_path
                orm.commit()


def setup(app):
    """ Register the CLI commands with the command parser """
    app.cli.add_command(publ_cli)



""" Full-text search stuff """
import datetime
import email
import logging
import os
import typing

import whoosh
import whoosh.fields
import whoosh.index
import whoosh.qparser
import whoosh.query
import whoosh.writing
from pony import orm
from werkzeug.utils import cached_property

from . import model, tokens, user, utils
from .entry import Entry

LOGGER = logging.getLogger(__name__)
SCHEMA_VERSION = 2


class SearchResults:
    """ Result set from a full-text search """

    def __init__(self, results):
        self._entries = [record
                         for record in [model.Entry.get(id=int(hit['entry_id']))
                                        for hit in results]
                         if record and record.visible]

    @cached_property
    def has_unauthorized(self):
        """ Returns whether there are any unauthorized entries present """

        cur_user = user.get_active()

        for record in self._entries:
            if not record.is_authorized(cur_user):
                tokens.request(cur_user)
                return True
        return False

    @cached_property
    def entries(self) -> typing.Callable[..., typing.List[Entry]]:
        """ Returns the entries in the search results """
        def _entries(unauthorized=0) -> typing.List[Entry]:
            return Entry.filter_auth(self._entries, unauthorized=unauthorized)

        return utils.CallableProxy(_entries)

    def __iter__(self):
        return self.entries().__iter__()

    def __len__(self):
        return len(self._entries)

    def __bool__(self):
        return bool(self._entries)


class SearchIndex:
    """ Full-text search index for all entries """

    def __init__(self, config):
        if not config.search_index:
            self.index = None
            return

        self.schema = whoosh.fields.Schema(
            entry_id=whoosh.fields.ID(stored=True, unique=True),
            title=whoosh.fields.TEXT,
            content=whoosh.fields.TEXT,
            published=whoosh.fields.DATETIME,
            tag=whoosh.fields.KEYWORD(lowercase=True, commas=True),
            category=whoosh.fields.ID,
            status=whoosh.fields.NUMERIC)

        if not os.path.exists(config.search_index):
            os.mkdir(config.search_index)

        with orm.db_session:
            version = model.GlobalConfig.get(key='search_index_version')

            if (not whoosh.index.exists_in(config.search_index)
                or not version
                    or version.int_value != SCHEMA_VERSION):
                # either the index doesn't exist or it's got the wrong schema;
                # (re)create it
                self.index = whoosh.index.create_in(config.search_index, self.schema)
                if version:
                    version.int_value = SCHEMA_VERSION
                else:
                    version = model.GlobalConfig(key='search_index_version',
                                                 int_value=SCHEMA_VERSION)
            else:
                self.index = whoosh.index.open_dir(config.search_index)

        self.query_parser = whoosh.qparser.QueryParser("content", self.index.schema)

    def update(self, record: model.Entry, entry_file: email.message.EmailMessage):
        """
        Add an entry to the content index
        """

        if not self.index:
            return

        if record.status not in (model.PublishStatus.PUBLISHED.value,
                                 model.PublishStatus.SCHEDULED.value):
            self.index.delete_by_term("entry_id", str(record.id))
            return

        with whoosh.writing.AsyncWriter(self.index) as writer:
            writer.update_document(
                entry_id=str(record.id),
                title=record.title,
                content=entry_file.get_payload(),
                published=datetime.datetime.fromtimestamp(record.utc_timestamp),
                tag=','.join(entry_file.get_all('tag') or []),
                category=record.category,
                status=record.status)

    def query(self, query: str,
              category=None, recurse=False,
              count=None, page=None,
              future=False):
        """
        Searches with a text query

        :param category: The category to search on; None to avoid restriction
        :param bool recurse: Whether to also match subcategories
        :param int count: The number of entries per page
        :param int page: The page number to retrieve
        :param bool future: Whether to retrieve entries that aren't yet visible
        """
        # pylint:disable=too-many-arguments
        LOGGER.debug('query: %s  category: %s  recurse: %s  future: %s',
                     query, category, recurse, future)

        if not self.index:
            return SearchResults([])

        with self.index.searcher() as searcher:
            parsed = self.query_parser.parse(query)

            if category is not None:
                if str(category):
                    # We're in a category other than root
                    cat_term = whoosh.query.Term("category", str(category))
                    if recurse:
                        cat_term = whoosh.query.Or([cat_term,
                                                    whoosh.query.Prefix("category",
                                                                        f"{str(category)}/")])
                    parsed = whoosh.query.And([parsed, cat_term])
                elif not recurse:
                    # We're in root and not recursing, so we need to match empty
                    parsed = whoosh.query.And([parsed, whoosh.query.Term("category", "")])

            if not future:
                parsed = whoosh.query.And([
                    parsed,
                    whoosh.query.Or([
                        whoosh.query.Term("status", model.PublishStatus.PUBLISHED.value),
                        whoosh.query.DateRange("published", None, datetime.datetime.now()),
                    ])
                ])

            LOGGER.debug('parse result: %s', parsed)
            if page is not None:
                results = searcher.search_page(parsed, page, pagelen=count)
            else:
                results = searcher.search(parsed, limit=count)

            return SearchResults(results)



# config.py
""" configuration container for Publ """

import logging
import os
import typing

import werkzeug.local
from dateutil import tz

LOGGER = logging.getLogger(__name__)


class _Defaults:
    # pylint:disable=too-few-public-methods
    database_config = {
        'provider': 'sqlite',
        'filename': ':memory:'
    }
    index_rescan_interval = 7200
    index_wait_time = 1
    index_enable_watchdog = True

    # Site content locations
    content_folder = 'content'
    template_folder = 'templates'
    static_folder = 'static'
    static_url_path = '/static'

    # Image rendition cache
    image_output_subdir = '_img'
    image_cache_interval = 3600
    image_cache_age = 86400 * 30  # one month
    image_render_threads = os.cpu_count()

    timezone = tz.tzlocal()

    # Page rendering
    cache: typing.Dict[str, str] = {}
    layout: typing.Dict[str, typing.Any] = {}
    markdown_extensions = (
        'tables',
        'fenced-code',
        'footnotes',
        'strikethrough',
        'highlight',
        'superscript',
        'math',
    )

    # Authentication
    auth: typing.Dict[str, str] = {}
    user_list = 'users.cfg'
    admin_group = 'admin'
    auth_log_prune_interval = 3600
    auth_log_prune_age = 86400 * 30  # one month
    ticket_lifetime = 60
    token_lifetime = 86400 * 30
    refresh_token_lifetime = 86400 * 90

    # Full-text search directory
    search_index = None


class Config(_Defaults):
    """ Stores configuration for a Publ app """
    # pylint:disable=too-few-public-methods

    def __init__(self, from_dict):
        # Copy over the defaults
        for key, val in _Defaults.__dict__.items():
            if key[0] != '_':
                setattr(self, key, val)

        # Copy over the new configuration
        for key, val in from_dict.items():
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                LOGGER.warning("Unknown configuration key %s", key)


def _get_current():
    from .flask_wrapper import current_app
    return current_app.publ_config


config = typing.cast(Config, werkzeug.local.LocalProxy(_get_current))  # pylint:disable=invalid-name



# cards.py
""" Rendering functions for Twitter/OpenGraph cards"""

import logging
import typing

from . import utils

LOGGER = logging.getLogger(__name__)


class CardData():
    """ Extracted card data """
    # pylint: disable=too-few-public-methods

    def __init__(self):
        self.images: typing.List[
            typing.Tuple[
                str,
                typing.Optional[str],
                typing.Optional[str]]] = []


class HtmlCardParser(utils.HTMLTransform):
    """ Parse the card data out of an HTML document """

    def __init__(self, card):
        super().__init__()

        self._card = card
        self._images = []

    def handle_starttag(self, tag, attrs):
        if tag == 'img':
            src = None
            width = None
            height = None
            for attr, val in attrs:
                if attr == 'src':
                    src = val
                elif attr == 'width':
                    width = val
                elif attr == 'height':
                    height = val
            if src:
                self._card.images.append((src, width, height))


def extract_card(html_text: str) -> CardData:
    """ Extract card data based on the provided HTML. """
    card = CardData()
    HtmlCardParser(card).feed(html_text)

    return card



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['publ', 'publ.image']

package_data = \
{'': ['*'], 'publ': ['default_template/*']}

install_requires = \
['Flask-Caching>=1.9.0,<2.0.0',
 'Flask>=2.0.0,<3.0.0',
 'Pillow>=9.0.0,<10.0.0',
 'Pygments>=2.7.3,<3.0.0',
 'Whoosh>=2.7.4,<3.0.0',
 'arrow>=0.17.0,<0.18.0',
 'atomicwrites>=1.4.0,<2.0.0',
 'authl>=0.6.0,<0.7.0',
 'awesome-slugify>=1.6.5,<2.0.0',
 'misaka>=2.1.1,<3.0.0',
 'pony>=0.7.14,<0.8.0',
 'watchdog>=1.0.2,<2.0.0']

setup_kwargs = {
    'name': 'publ',
    'version': '0.7.8',
    'description': 'A flexible web-based publishing framework',
    'long_description': "# Publ\n\nA personal publishing platform. Like a static publishing system, only dynamic.\n\n## Motivation\n\nI make a lot of different things — comics, music, art, code, games — and none of\nthe existing content management systems I found quite satisfied my use cases.\nEither they don't allow enough flexibility in the sorts of content that they can\nprovide, or the complexity in managing the content makes it more complicated than\nsimply hand-authoring a site.\n\nI wanted to bring the best of the classic static web to a more dynamic\npublishing system; scheduled posts, private posts, category-based templates, and\nbuilt-in support for image renditions (including thumbnails, high-DPI support,\nand image galleries). And I want to do it all using simple Markdown files\norganized in a sensible file hierarchy.\n\n## Basic tenets\n\n* Containerized web app that's deployable with little friction (hopefully)\n* Do one thing (present heterogeneous content), do it well (hopefully)\n* Use external tools for site content editing\n* Be CDN-friendly\n* High-DPI images and image sets as first-class citizens\n* Interoperate with everything that's open for interoperation (especially [IndieWeb](http://indieweb.org))\n\n## See it in action\n\nThe main demonstration site is at https://beesbuzz.biz/ — it is of course a\nwork in progress! The documentation site for Publ itself (which is also a work in progress) lives at https://publ.plaidweb.site/\n\n## Operating requirements\n\nI am designing this to work in any WSGI-capable environment with a supported\nversion of Python. This means that it will, for example, be deployable on any\nshared hosting which has Passenger support (such as Dreamhost), as well as on\nHeroku, Google AppEngine, S3, or any other simple containerized deployment\ntarget.\n\nThe file system is the ground truth for all site data, and while it does use a\ndatabase as a content index, the actual choice of database doesn't matter all\nthat much. A typical deployment will use SQLite, but MySQL, Postgres, Oracle,\nand Cockroach are also supported.\n\n## Developing Publ\n\nIn order to develop Publ itself, you'll need to install its dependencies; see\nthe [getting started\nguide](http://publ.plaidweb.site/manual/328-Getting-started) for more\ninformation. In particular, make sure you have compatible versions of\n[Python](https://python.org/) and [Poetry](https://python-poetry.org/)\ninstalled, and, if on Windows, you'll probably need to install the [Visual C++\nbuild tools](https://visualstudio.microsoft.com/downloads/).\n\nAs far as developing Publ itself goes, cloning this repository and running\n`./runTests.sh` (Linux/macOS/etc.) or `wintests.cmd` (Windows) should get you up\nand running. The runtime manual test suite site lives in `tests/` (with the\nactual site content in `content/`, `templates/` and `static/`).\n\nFor developing CLI functionality, you'll have to override the `FLASK_APP`\nenvironment variable to be `test_app.py`.\n\n## Additional resources\n\nThe [Publ-site](https://github.com/PlaidWeb/Publ-site) repository stores all of\nthe templates, site content, and configuration for the [Publ\nsite](https://publ.plaidweb.site).\n\nThe\n[Publ-templates-beesbuzz.biz](https://github.com/PlaidWeb/Publ-templates-beesbuzz.biz)\nrepository provides a stripped-down sample site based on [my personal\nhomepage](https://beesbuzz.biz).\n\n## Authors\n\nIn order of first contribution:\n\n* [fluffy](https://github.com/fluffy-critter)\n* [karinassuni](https://github.com/karinassuni)\n",
    'author': 'fluffy',
    'author_email': 'fluffy@beesbuzz.biz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://publ.plaidweb.site/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# caching.py
""" Useful caching functions """

import hashlib
from abc import ABC, abstractmethod

from flask import request
from flask_caching import Cache

cache = Cache()  # pylint: disable=invalid-name


def init_app(app, config):
    """ Initialize the cache for the app """
    cache.init_app(app, config=config)


def do_not_cache():
    """ Return whether we should cache a page render """

    from . import index  # pylint: disable=cyclic-import

    if index.in_progress():
        # We are reindexing the site
        return True

    return False


def not_modified(etag):
    """ Return True if the request indicates that the client's cache is valid """

    if request.if_none_match.contains(etag):
        return True

    return False


def get_etag(text):
    """ Compute the etag for the rendered text"""

    return hashlib.md5(text.encode('utf-8')).hexdigest()


class Memoizable(ABC):
    """ Add this interface to a class to make it stably memoizable. """

    @abstractmethod
    def _key(self):
        """
            This should return a value that will be unique across all
            objects of this class.
        """

    def __repr__(self):
        clss = self.__class__.__name__
        key = self._key()
        return f"{clss}({key})".replace(' ', '_')

    def __hash__(self):
        return hash(self._key())

    def __eq__(self, other):
        # pylint: disable=protected-access
        return self._key() == other._key()



# utils.py
""" Some useful utilities that don't belong anywhere else """

import ast
import functools
import html
import html.parser
import logging
import os
import re
import typing
import urllib.parse

import arrow
import flask
import slugify
import werkzeug.routing

from . import model
from .config import config

LOGGER = logging.getLogger(__name__)

T = typing.TypeVar('T')  # pylint:disable=invalid-name
ArgDict = typing.Dict[str, typing.Any]
ListLike = typing.Union[typing.List[T],
                        typing.Tuple[T, ...],
                        typing.Set[T]]
TagAttr = typing.Union[str, bool, None]
TagAttrs = typing.Dict[str, TagAttr]

_TagKey = slugify.Slugify(to_lower=True)  # type:ignore


class CallableProxy:
    """ Wrapper class to make args possible on properties. """
    # pylint: disable=too-few-public-methods

    def __init__(self, func: typing.Callable[..., T]):
        """ Construct the property proxy.

        :param func: The function to wrap
        """

        self._func: typing.Callable[..., T] = func

    @functools.lru_cache()
    def _cached_default(self, *_):
        """ caching wrapper to memoize call against everything that might affect it """
        return self._func()

    def _default(self):
        from . import user  # pylint:disable=cyclic-import
        return self._cached_default(flask.request.url, user.get_active())

    def __call__(self, *args, **kwargs) -> T:
        # use the new kwargs to override the defaults
        return self._func(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self._default(), name)

    def __bool__(self) -> bool:
        return bool(self._default())

    def __len__(self) -> int:
        return len(self._default())

    def __str__(self) -> str:
        return str(self._default())

    def __iter__(self):
        return self._default().__iter__()

    def __getitem__(self, key):
        return self._default().__getitem__(key)

    def __hash__(self):
        return hash((CallableProxy, self._func))

    def __eq__(self, other):
        return self._default() == other

    def __contains__(self, item):
        return item in self._default()

    def __add__(self, other):
        return self._default() + other


class TrueCallableProxy(CallableProxy):
    """ A version of CallableProxy that is always truthy """
    # pylint: disable=too-few-public-methods

    def __bool__(self):
        return True


class CallableValue(CallableProxy):
    """ A version of CallableProxy that returns a fixed value """
    # pylint:disable=too-few-public-methods

    def __init__(self, value):
        super().__init__(lambda *args, **kwargs: value)


#: arrow format string for 'day' archives
DAY_FORMAT = 'YYYY-MM-DD'

#: arrow format string for 'month' archives
MONTH_FORMAT = 'YYYY-MM'

#: arrow format string for 'year' archives
YEAR_FORMAT = 'YYYY'

#: arrow format string for 'week' archives
WEEK_FORMAT = 'YYYYMMDD_w'


def parse_date(datestr: str) -> typing.Tuple[arrow.Arrow, str, str]:
    """ Parse a date expression into a tuple of:
    ``(start_date, span_type, span_format)``

    :param str datestr: A date specification, in the format of YYYY-MM-DD (dashes optional)
    """

    match = re.match(
        r'([0-9]{4})(-?([0-9]{1,2}))?(-?([0-9]{1,2}))?(_w)?$', datestr)
    if not match:
        return (arrow.get(datestr,
                          tzinfo=config.timezone).replace(tzinfo=config.timezone),
                'day', DAY_FORMAT)

    year, month, day, week = match.group(1, 3, 5, 6)
    start = arrow.Arrow(year=int(year), month=int(
        month or 1), day=int(day or 1), tzinfo=config.timezone)

    if week:
        return start.span('week')[0], 'week', WEEK_FORMAT

    if day:
        return start, 'day', DAY_FORMAT

    if month:
        return start, 'month', MONTH_FORMAT

    return start, 'year', YEAR_FORMAT


def find_file(path: str, search_path: typing.Union[str, ListLike[str]]) -> typing.Optional[str]:
    """ Find a file by relative path. Arguments:

    :param str path: the image's filename
    :param list search_path: a list of directories to check in

    :returns: the resolved file path
    """

    for relative in as_list(search_path):
        candidate = os.path.normpath(os.path.join(relative, path))
        if os.path.isfile(candidate):
            return candidate

    return None


def static_url(path: str, absolute: bool = False) -> str:
    """ Shorthand for returning a URL for the requested static file.

    Arguments:

    :param str path: the path to the file (relative to the static files directory)
    :param bool absolute: whether the link should be absolute or relative
    """

    if os.sep != '/':
        path = '/'.join(path.split(os.sep))

    return flask.url_for('static', filename=path, _external=absolute)


def make_tag(name: str,
             attrs: TagAttrs,
             start_end: bool = False) -> str:
    """ Build an HTML tag from the given name and attributes.

    :param str name: the name of the tag (p, div, etc.)
    :param attrs: a dict or list of attributes to apply to the tag
    :param bool start_end: whether this tag should be self-closing

    If an attribute's value is None it will be written as a standalone attribute,
    e.g. ``<audio controls>``. To suppress it entirely, make the value explicitly False.
    """

    text = '<' + name

    if isinstance(attrs, dict):
        attr_list = attrs.items()
    elif isinstance(attrs, list):
        attr_list = attrs
    elif attrs is not None:
        raise TypeError("Unhandled attrs type " + str(type(attrs)))

    for key, val in attr_list:
        if val is not False:
            text += f' {key}'
            if val is not None:
                import markupsafe
                escaped = html.escape(str(val), False).replace('"', '&#34;')

                if isinstance(val, CallableProxy):
                    val = val()
                if isinstance(val, markupsafe.Markup):
                    # We just double-escaped all entities...
                    escaped = re.sub(r'&amp;([a-zA-Z0-9.\-_\:]+;)', r'&\1', val)
                text += f'="{escaped}"'
    if start_end:
        text += ' /' if attrs else '/'
    text += '>'
    return text


def file_fingerprint(fullpath: str) -> str:
    """ Get a metadata fingerprint for a file """
    try:
        stat = os.stat(fullpath)
        return ','.join([str(value)
                         for value in [stat.st_ino, stat.st_mtime, stat.st_size]
                         if value])
    except FileNotFoundError:
        LOGGER.warning("Attempted to get fingerprint of nonexistent file %s", fullpath)
        return ''


def remap_args(input_args: typing.Dict[str, typing.Any],
               remap: typing.Dict[str, typing.Union[str, ListLike[str]]]
               ) -> typing.Dict[str, typing.Any]:
    """ Generate a new argument list by remapping keys.

    :param dict input_args: The input argument list
    :param dict remap: A mapping of destination key -> priority list of source keys
    """
    out_args = input_args
    for dest_key, src_keys in remap.items():
        remap_value = None

        for key in as_list(src_keys):
            if key in input_args:
                remap_value = input_args[key]
                break

        if remap_value is not None:
            if out_args is input_args:
                out_args = {**input_args}
            out_args[dest_key] = remap_value

    return out_args


def remap_link_target(path: str, absolute: bool = False) -> str:
    """ Remap a link target from a path, URL, or static file reference.

    :param str path: The original link target; start with ``@`` to indicate a
        static file

    :param bool absolute: Whether to make the link absolute with respect to the
        current page URL
    """
    if path.startswith('@'):
        # static resource
        return static_url(path[1:], absolute=absolute)

    if absolute:
        # absolute-ify whatever the URL is
        return urllib.parse.urljoin(flask.request.url, path)

    return path


def get_category(filename: str) -> str:
    """ Get a default category name from a filename in a cross-platform manner """
    return '/'.join(os.path.dirname(filename).split(os.sep))


class HTMLTransform(html.parser.HTMLParser):
    """ Wrapper to HTMLParser to make it easier to build a SAX-style processor.

    You will probably want to implement:

    * ``handle_starttag(self, tag, attrs)``
    * ``handle_endtag(self, tag)``
    * ``handle_data(self, data)``
    * ``handle_startendtag(self, tag, attrs)``
    """

    def __init__(self):
        super().__init__()

        self.reset()
        self.strict = False
        self.convert_charrefs = False
        self._fed = []

    def feed(self, data: str):
        """ Feed in some text data.

        Overrides the base class to ensure that it's handled like a plain string
        and not a MarkupSafe string (which causes double-escaping to happen)
        """
        super().feed(str(data))

    def append(self, item: str):
        """ Append some text to the output """
        self._fed.append(item)

    def get_data(self) -> str:
        """ Concatenate the output """
        return ''.join(self._fed)

    def handle_entityref(self, name: str):
        self.handle_data('&' + name + ';')

    def handle_charref(self, name: str):
        self.handle_data('&#' + name + ';')

    def error(self, message: str):
        """ Deprecated, per https://bugs.python.org/issue31844 """


def prefix_normalize(kwargs: ArgDict) -> ArgDict:
    """

    Given an argument list where one of them is ``prefix``, normalize the
    arguments to convert ``{prefix}{key}`` to ``{key}`` and remove the prefixed
    versions

    For example::

        {
            'prefix': 'index_',
            'width': 320,
            'index_width': 256,
        }

    becomes::

        {
            'width': 256
        }

    """

    prefixed = {}
    removed = []
    if 'prefix' in kwargs:
        for prefix in as_list(kwargs.get('prefix')):
            for k, val in kwargs.items():
                if k.startswith(prefix):
                    prefixed[k[len(prefix):]] = val
                    removed.append(k)

    normalized = {**kwargs, **prefixed}
    for k in removed:
        normalized.pop(k, None)
    return normalized


def is_list(item: typing.Any) -> bool:
    """ Return if this is a list-type thing """
    return bool(getattr(item, '__iter__', None)) and not isinstance(item, str)


def as_list(item: typing.Any) -> ListLike:
    """ Return list-type things directly; convert other things into a tuple """
    if item is None:
        return ()

    if is_list(item):
        return item

    return (item,)


class CategoryConverter(werkzeug.routing.PathConverter):
    """ A version of PathConverter that only accepts valid Publ categories
    (i.e. no segment starts with _) """

    def to_python(self, value: str) -> str:
        for part in value.split('/'):
            if part[0] == '_':
                raise werkzeug.routing.ValidationError
        return super().to_python(value)


class TemplateConverter(werkzeug.routing.UnicodeConverter):
    """ A version of UnicodeConverter that doesn't accept strings beginning with _ """

    def to_python(self, value: str) -> str:
        if value[0] == '_':
            raise werkzeug.routing.ValidationError
        return super().to_python(value)


def redir_path(path: str = None) -> str:
    """ Convert a URI path to a path fragment, suitable for url_for

    :param str path: The path to redirect to; uses the current request.full_path
        if unspecified

    """

    if path is None:
        path = flask.request.full_path
    else:
        path = str(path)

    # strip off leading slashes
    path = re.sub(r'^/*', r'', path)
    # if there's a trailing ? strip that off too
    path = re.sub(r'\?$', r'', path)

    return path


def secure_link(endpoint: str, *args, **kwargs) -> str:
    """ :py:func:`flask.url`_for except it will force the link to be secure if
    we are configured with ``AUTH_FORCE_HTTPS`` """
    force_ssl = config.auth.get('AUTH_FORCE_HTTPS')

    if force_ssl and flask.request.scheme != 'https':
        kwargs = {**kwargs,
                  '_external': True,
                  '_scheme': 'https'}
    return flask.url_for(endpoint, *args, **kwargs)


def auth_link(endpoint: str, auto_redir=True) -> typing.Callable[..., str]:
    """ Generates a function that maps an optional redir parameter to the
    specified endpoint. If redir is unspecified it defaults to the current
    request's full_path. """
    def endpoint_link(redir=None, absolute=False, **kwargs):
        LOGGER.debug("Getting %s for redir=%s kwargs=%s", endpoint, redir, kwargs)
        redir = redir_path(redir) if redir or auto_redir else None

        return secure_link(endpoint, redir=redir, _external=absolute, **kwargs)

    return CallableProxy(endpoint_link)


def stash(func: typing.Optional[typing.Callable] = None):
    """ Decorator to memoize a function onto the request's global context.
    """

    def make_hashable(item):
        from . import caching
        if isinstance(item, (str, caching.Memoizable)):
            return item
        if isinstance(item, (list, tuple)):
            return tuple(make_hashable(i) for i in item)
        if isinstance(item, set):
            return frozenset(make_hashable(i) for i in item)
        if isinstance(item, dict):
            return frozenset((make_hashable(k), make_hashable(v)) for k, v in item.items())
        return item

    def decorator(inner: typing.Callable):
        def wrapped_func(*args, **kwargs):
            if 'store' not in flask.g:
                flask.g.store = {}  # pylint:disable=assigning-non-slot
            if inner not in flask.g.store:
                flask.g.store[inner] = {}
            store = flask.g.store[inner]

            cache_key = (make_hashable(args), make_hashable(kwargs))
            if cache_key in store:
                return store[cache_key]

            val = inner(*args, **kwargs)
            store[cache_key] = val
            return val
        return wrapped_func

    return decorator(func) if func else decorator


def parse_tuple_string(argument: typing.Union[str, typing.Tuple, typing.List],
                       type_func=int) -> typing.Optional[typing.Tuple]:
    """ Return a tuple from parsing 'a,b,c,d' -> (a,b,c,d) """
    if argument is None:
        return None
    if isinstance(argument, str):
        return tuple(type_func(p.strip()) for p in argument.split(','))
    return tuple(argument)


@functools.lru_cache()
def tag_key(tag):
    """ Given a tag, return its normalized key """
    if isinstance(tag, str):
        return _TagKey(tag)
    if isinstance(tag, model.EntryTag):
        return tag.key
    if isinstance(tag, model.EntryTagged):
        return tag.tag.key
    raise ValueError(f"TagSet got non-tag-type {type(tag)}")


def tag_cname(tag) -> str:
    """ Get the canonical name of a tag from a name or key """
    record = model.EntryTag.get(key=tag_key(tag))
    if record:
        return record.name
    return tag


class TagSet(typing.Set[str]):
    """ A frozenset-equivalent class that is case-insensitive """

    def __init__(self, contents: ListLike[str] = None):
        super().__init__()
        if contents:
            storage = {tag_key(v): tag_cname(v) for v in contents}
            self._keys = frozenset(storage.keys())
            self._values = frozenset(storage.values())
        else:
            self._keys = frozenset()
            self._values = frozenset()

    def __contains__(self, key) -> bool:
        return tag_key(key) in self._keys

    def __iter__(self):
        return self._values.__iter__()

    def __hash__(self):
        return hash(self._keys)

    def __repr__(self):
        return f'{self.__class__}({set(self._values)})'

    def __str__(self):
        return str(set(self._values))

    def __or__(self, other):
        return TagSet(list(self) + list(other))

    @staticmethod
    def _fold(items):
        return {tag_key(k) for k in items}

    def __and__(self, other):
        folded = self._fold(other)
        return TagSet((k for k in self if tag_key(k) in folded))

    def __xor__(self, other):
        folded = self._fold(self) ^ self._fold(other)
        return TagSet((k for k in self | other if tag_key(k) in folded))

    def __sub__(self, other):
        folded = self._fold(other)
        return TagSet((k for k in self if tag_key(k) not in folded))

    def __len__(self):
        return len(self._values)

    def __bool__(self):
        return bool(self._values)

    def __eq__(self, other):
        folded = self._fold(other)
        return self._keys == folded

    def __ne__(self, other):
        folded = self._fold(other)
        return self._keys != folded

    def __le__(self, other):
        return self._fold(self) <= self._fold(other)

    def __lt__(self, other):
        return self._fold(self) < self._fold(other)

    def keys(self):
        """ Return the case-folded names """
        return self._keys

    def values(self):
        """ return the case-preserved names """
        return self._values


def strip_single_paragraph(text: str):
    """ If the provided HTML text has only a single paragraph, strip it off. """

    stripped = re.sub(r'^<p>(.*)</p>$', r'\1', text.strip())
    if '<p>' in stripped:
        return text
    return stripped


def parse_spec(text: str, pos_limit: int = None) -> typing.Tuple[str, list, ArgDict]:
    """ Given a string like ``foo{10,bar=baz}``, parse out the argument lists.

    :param str text: The text to parse
    :param int pos_limit: The maximum number of positional args to allow
    :returns: Tuple of text, pos_args, kw_args
    """
    match = re.match(r'([^\{]*)(\{(.*)\})$', text)
    if match:
        pos_args, kw_args = parse_arglist(match.group(3), pos_limit)
        return match.group(1), pos_args, kw_args

    return text, [], {}


def parse_arglist(args: str, pos_limit: int = None) -> typing.Tuple[list, ArgDict]:
    """ Parse an argument list into pos_args, kw_args"""
    tree = ast.parse(f'f({args})')
    expr = typing.cast(ast.Expr, tree.body[0])
    funccall = typing.cast(ast.Call, expr.value)

    pos_args = [ast.literal_eval(arg) for arg in funccall.args]
    if pos_limit is not None and len(pos_args) > pos_limit:
        raise TypeError(
            f"Expected at most {pos_limit} positional args but {len(pos_args)} were given")

    kwargs = {arg.arg: ast.literal_eval(arg.value)
              for arg in funccall.keywords if arg.arg}

    LOGGER.debug("pos_args=%s kw_args=%s", pos_args, kwargs)
    return pos_args, kwargs


def canonicize_url(url: str) -> str:
    """ Canonicize a URL to make them string-comparable """
    assert url is not None

    parsed = urllib.parse.urlparse(url)._asdict()
    parsed['netloc'] = parsed['netloc'].casefold()
    if not parsed.get('path'):
        parsed['path'] = '/'
    return urllib.parse.urlunparse(urllib.parse.ParseResult(**parsed))



""" Functionality for images external to the content directory """

import html
import os.path
import urllib.parse
from abc import abstractmethod

import flask

from .image import Image


class ExternalImage(Image):
    """ Base class for images which are rendered by external means """

    @abstractmethod
    def _get_url(self, absolute):
        """ Implemented by subclasses to actually map the URL """

    CSS_SIZE_MODE = {
        'fit': 'contain',
        'fill': 'cover'
    }

    def get_rendition(self, output_scale=1, **kwargs):
        # pylint: disable=unused-argument
        return self._get_url(kwargs.get('absolute')), None

    def _get_img_attrs(self, spec, style_parts):
        url = self._get_url(spec.get('absolute'))

        attrs = {'src': url}
        if 'class' in spec or 'img_class' in spec:
            attrs['class'] = spec.get('class', spec.get('img_class'))
        if 'id' in spec:
            attrs['id'] = spec['id']

        if not spec.get('_no_resize_external'):
            # try to fudge the sizing
            max_width = spec.get('max_width')
            width = spec.get('width')
            max_height = spec.get('max_height')
            height = spec.get('height')
            size_mode = spec.get('resize', 'fit')

            if width and max_width and max_width < width:
                if height:
                    height = height * max_width / width
                width = max_width
            if height and max_height and max_height < height:
                if width:
                    width = width * max_height / height
                height = max_height

            if width and height and size_mode != 'stretch':
                fill_crop_x = spec.get('fill_crop_x', 0.5) * 100
                fill_crop_y = spec.get('fill_crop_y', 0.5) * 100
                style_parts += [
                    f'background-image:url(\'{html.escape(url)}\')',
                    f'background-size:{self.CSS_SIZE_MODE[size_mode]}',
                    f'background-position:{fill_crop_x:.2f}% {fill_crop_y:.2f}%',
                    'background-repeat:no-repeat'
                ]
                attrs['src'] = flask.url_for(
                    'chit', _external=spec.get('absolute'))

            if width:
                attrs['width'] = width
            if height:
                attrs['height'] = height
            if max_width:
                style_parts.append(f'max-width:{max_width}px')
            if max_height:
                style_parts.append(f'max-height:{max_height}px')

        return attrs

    def _css_background(self, **kwargs) -> str:
        """ Get the CSS background-image for the remote image """
        return f'background-image: url("{self._get_url(kwargs.get("absolute"))}");'

    @property
    def _filename(self) -> str:
        return os.path.basename(urllib.parse.urlparse(self._get_url(True)).path)



# category.py
""" The Category object passed to entry and category views """

import collections
import email
import logging
import os
import typing

import flask
from flask import url_for
from pony import orm
from werkzeug.utils import cached_property

from . import entry  # pylint: disable=cyclic-import
from . import caching, markdown, model, path_alias, queries, utils
from .config import config

LOGGER = logging.getLogger(__name__)

TagCount = collections.namedtuple('TagCount', ['name', 'count'])


def load_metafile(filepath):
    """ Load a metadata file from the filesystem """
    try:
        with open(filepath, 'r', encoding='utf-8') as file:
            return email.message_from_file(file)
    except FileNotFoundError:
        LOGGER.warning("Category file %s not found", filepath)

        # SQLite doesn't support cascading deletes, so clean up
        orm.delete(pa for pa in model.PathAlias if pa.category.file_path == filepath)

        orm.delete(c for c in model.Category if c.file_path == filepath)
        orm.commit()

    return None


def search_path(category: str) -> str:
    """ Return the file search path for a named category """
    return os.path.join(config.content_folder, category)


class Category(caching.Memoizable):
    """ Wrapper for category information """
    # pylint: disable=too-few-public-methods,too-many-instance-attributes

    __hash__ = caching.Memoizable.__hash__  # type:ignore

    @staticmethod
    @utils.stash
    def load(path: str):
        """ Get a category wrapper object

        path -- the path to the category
        """
        return Category(Category.load.__name__, path)

    def __init__(self, create_key, path: str):
        """ Initialize a category wrapper """

        assert create_key == Category.load.__name__, "Category must be created with Category.load()"

        if path is None:
            path = ''

        self.path = path
        self.basename = os.path.basename(path)

        subcat_query = orm.select(e.category for e in model.Entry if e.visible)  # type:ignore
        if path:
            subcat_query = orm.select(
                c for c in subcat_query if c.startswith(path + '/'))
        else:
            subcat_query = orm.select(c for c in subcat_query if c != '')
        self._subcats_recursive = subcat_query

        self._record = model.Category.get(category=path)

    def _key(self):
        return self.path

    @cached_property
    def link(self) -> typing.Callable[..., str]:
        """ Returns a link to the category.

        Takes optional view arguments, as well as the following optional arguments:

        template -- which template to generate the link against

        """
        def _link(template='', absolute=False, **kwargs) -> str:
            return url_for('category',
                           category=self.path,
                           template=template,
                           _external=absolute,
                           **kwargs)

        return utils.CallableProxy(_link)

    @cached_property
    def subcats(self) -> typing.Callable[..., str]:
        """ Returns a list of subcategories.

        Takes the following arguments:

        recurse -- whether to include their subcategories as well (default: False)

        """
        def _get_subcats(recurse=False) -> typing.List["Category"]:
            """ Get the subcategories of this category

            recurse -- whether to include their subcategories as well
            """

            if recurse:
                # No need to filter
                return sorted([Category.load(e) for e in self._subcats_recursive],
                              key=lambda cat: tuple(crumb.sort_name for crumb in cat.breadcrumb))

            # get all the subcategories, with only the first subdir added

            # number of path components to ingest
            parts = len(self.path.split('/')) + 1 if self.path else 1

            # convert the subcategories into separated pathlists with only 'parts'
            # parts, then rejoin back into a path and make unique
            subcats = {'/'.join(c.split('/', parts)[:parts]) for c in self._subcats_recursive}

            # convert to a bunch of Category objects
            return sorted([Category.load(c) for c in subcats], key=lambda c: c.sort_name)

        return utils.CallableProxy(_get_subcats)

    @cached_property
    def first(self) -> typing.Callable[..., typing.Optional[entry.Entry]]:
        """ Returns the first entry in the category.

        Takes optional view arguments.
        """
        def _first(**spec) -> typing.Optional[entry.Entry]:
            """ Get the earliest entry in this category, optionally including subcategories """
            for record in self._entries(spec).order_by(model.Entry.local_date,
                                                       model.Entry.id)[:1]:
                return entry.Entry.load(record)
            return None

        return utils.CallableProxy(_first)

    @cached_property
    def last(self) -> typing.Callable[..., typing.Optional[entry.Entry]]:
        """ Returns the last entry in the category.

        Takes optional view arguments.
        """
        def _last(**spec) -> typing.Optional[entry.Entry]:
            """ Get the latest entry in this category, optionally including subcategories """
            for record in self._entries(spec).order_by(orm.desc(model.Entry.local_date),
                                                       orm.desc(model.Entry.id))[:1]:
                return entry.Entry.load(record)
            return None

        return utils.CallableProxy(_last)

    @cached_property
    def _meta(self) -> typing.Optional[email.message.Message]:
        if self._record and self._record.file_path:
            return load_metafile(self._record.file_path)

        return None

    @cached_property
    def name(self) -> typing.Callable[..., str]:
        """ Get the display name of the category. Accepts the following arguments:

        markup -- If True, convert it from Markdown to HTML; otherwise, strip
            all markup (default: True)
        no_smartquotes -- if True, preserve quotes and other characters as originally
            presented
        markdown_extensions -- a list of markdown extensions to use
        """
        if self._meta and self._meta.get('name'):
            # get it from the meta file
            name = self._meta.get('name')
        else:
            # infer it from the basename
            name = self.basename.replace('_', ' ').title()

        def _name(markup=True, no_smartquotes=False, markdown_extensions=None) -> str:
            return markdown.render_title(name, markup, no_smartquotes,
                                         markdown_extensions)
        return utils.CallableProxy(_name)

    @cached_property
    def description(self) -> typing.Callable[..., str]:
        """ Get the textual description of the category """
        def _description(**kwargs) -> str:
            if self._meta:
                return flask.Markup(markdown.to_html(self._meta.get_payload(),
                                                     args=kwargs,
                                                     search_path=self.search_path,
                                                     counter=markdown.ItemCounter()))
            return ''

        if self._meta and self._meta.get_payload():
            return utils.TrueCallableProxy(_description)
        return utils.CallableValue('')

    @cached_property
    def search_path(self) -> str:
        """ Get the file search path for this category """
        return search_path(self.path)

    @cached_property
    def breadcrumb(self) -> typing.List["Category"]:
        """ Get the category hierarchy leading up to this category, including
        root and self.

        For example, path/to/long/category will return a list containing
        Category.load('path'), Category.load('path/to'), and Category.load('path/to/long').
        """
        ret = []
        here: typing.Optional[Category] = self
        while here:
            ret.append(here)
            here = here.parent
        return list(reversed(ret))

    @cached_property
    def sort_name(self) -> str:
        """ Get the sorting name of this category """
        if self._record and self._record.sort_name:
            return self._record.sort_name
        return self.name(markup=False)

    @cached_property
    def tags(self) -> typing.Callable[..., typing.List[TagCount]]:
        """ Get the list of tags associated with this category's entries.

        Takes optional view arguments (including recurse)

        Returns a list of category.TagCount tuples like `(tag='foo', count=5)`
        """
        def _tags(**spec) -> typing.List[TagCount]:
            entries = self._entries(spec)
            etags = orm.select((tag.tag, orm.count(tag.tag, distinct=False))
                               for e in entries for tag in e.tags if not tag.hidden)
            return [TagCount(tag.name, count) for (tag, count) in etags]
        return utils.CallableProxy(_tags)

    def __getattr__(self, name):
        """ Proxy undefined properties to the meta file """
        if self._meta:
            return self._meta.get(name)
        return None

    def get(self, name):
        """ Get a single metadata value """
        if self._meta:
            return self._meta.get(name)
        return None

    def get_all(self, name):
        """ Get all matching metadata values """
        if self._meta:
            return self._meta.get_all(name) or []
        return None

    def __str__(self):
        return self.path

    def __eq__(self, other):
        return str(other) == str(self)

    def __lt__(self, other):
        return self.path < other.path

    @cached_property
    def parent(self) -> typing.Optional["Category"]:
        """ Get the parent category """
        if self.path:
            return Category.load(os.path.dirname(self.path))
        return None

    @cached_property
    def root(self):
        """ Get the root category object. Equivalent to `breadcrumb[0]` but faster/easier. """
        if self.path:
            return Category.load(None)
        return self

    def _entries(self, spec):
        """ Return a model query to get our entry records """
        return queries.build_query({**spec, 'category': self})


@orm.db_session(retry=5)
def scan_file(fullpath, relpath) -> bool:
    """ scan a file and put it into the index """

    meta = load_metafile(fullpath)
    if not meta:
        return True

    # update the category meta file mapping
    category = meta.get('Category', utils.get_category(relpath))
    values = {
        'category': category,
        'file_path': fullpath,
        'sort_name': meta.get('Sort-Name', '')
    }

    LOGGER.debug("setting category %s to metafile %s", category, fullpath)
    record = model.Category.get(category=category)
    if record:
        record.set(**values)
    else:
        record = model.Category(**values)

    # update other relationships to the index
    path_alias.remove_aliases(record)
    for alias in meta.get_all('Path-Alias', []):
        path_alias.set_alias(alias, model.AliasType.REDIRECT, category=record)
    for alias in meta.get_all('Path-Mount', []):
        path_alias.set_alias(alias, model.AliasType.MOUNT, category=record)

    orm.commit()

    return True
